@@ -33,7 +33,7 @@ from ._tango import DevFailed, GreenMode, SerialModel
 from .attr_data import AttrData
 from .pipe_data import PipeData
 from .device_class import DeviceClass
-from .device_server import LatestDeviceImpl
+from .device_server import LatestDeviceImpl, get_worker, set_worker
 from .utils import get_enum_labels
 from .utils import is_seq, is_non_str_seq
 from .utils import scalar_to_array_type, TO_TANGO_TYPE
@@ -46,20 +46,6 @@ __all__ = ("DeviceMeta", "Device", "LatestDeviceImpl", "attribute",
 
 API_VERSION = 2
 
-# Worker access
-
-_WORKER = get_executor()
-
-
-def set_worker(worker):
-    global _WORKER
-    _WORKER = worker
-
-
-def get_worker():
-    return _WORKER
-
-
 # Helpers
 
 def _get_tango_type_format(dtype=None, dformat=None):
@@ -428,9 +414,10 @@ class _DeviceClass(DeviceClass):
             if init_dyn_attrs and callable(init_dyn_attrs):
                 try:
                     init_dyn_attrs()
-                except Exception:
+                except Exception as ex:
                     dev.warn_stream("Failed to initialize dynamic attributes")
                     dev.debug_stream("Details: " + traceback.format_exc())
+                    raise Exception(repr(ex))
 
 
 def __create_tango_deviceclass_klass(tango_device_klass, attrs=None):
@@ -514,7 +501,7 @@ def is_tango_object(arg):
     """Return tango data if the argument is a tango object,
     False otherwise.
     """
-    classes = attribute, device_property
+    classes = attribute, device_property, pipe
     if isinstance(arg, classes):
         return arg
     try:
@@ -1212,9 +1199,10 @@ class device_property(_BaseProperty):
                                               update_db)
         self.mandatory = mandatory
         if mandatory and default_value is not None:
-            msg = "device_property arguments mandatory " \
-                  "and default_value are incompatible"
-            raise Exception(msg)
+            msg = ("Invalid arguments: 'mandatory' is True, so 'default_value' must be None. "
+                   "A mandatory device property value must be defined in the Tango Database "
+                   "so it cannot have a default.")
+            raise ValueError(msg)
 
 
 class class_property(_BaseProperty):
@@ -1314,11 +1302,40 @@ def _add_classes(util, classes):
         util.add_class(*class_info)
 
 
+def _get_green_mode_from_classes_or_global_setting(classes):
+    green_modes = set()
+    global_green_mode = get_green_mode()
+    for _, klass, _ in _to_classes(classes):
+        device_green_mode = getattr(klass, 'green_mode', None)
+        if device_green_mode is None:
+            device_green_mode = global_green_mode
+        green_modes.add(device_green_mode)
+    if len(green_modes) > 1:
+        raise ValueError(
+            "Devices with mixed green modes cannot be run in the same device "
+            "server process. Modes: {}. Classes: {}.".format(green_modes, classes)
+        )
+    elif len(green_modes) == 0:
+        raise ValueError("No device classes specified - cannot run device server " 
+                         "process with no classes.")
+    unanimous_green_mode = green_modes.pop()
+    return unanimous_green_mode
+
+
 def __server_run(classes, args=None, msg_stream=sys.stdout, util=None,
                  event_loop=None, post_init_callback=None,
                  green_mode=None):
+    device_green_mode = _get_green_mode_from_classes_or_global_setting(classes)
     if green_mode is None:
-        green_mode = get_green_mode()
+        green_mode = device_green_mode
+    else:
+        if device_green_mode != green_mode:
+            raise ValueError(
+                "Invalid request for mixed green mode in the same "
+                "device server process. Device's mode: {} != kwarg mode: {}.".format(
+                    device_green_mode, green_mode
+                )
+            )
 
     write = msg_stream.write if msg_stream else lambda msg: None
 
@@ -1399,7 +1416,7 @@ def run(classes, args=None, msg_stream=sys.stdout,
     Example 1: registering and running a PowerSupply inheriting from
     :class:`~tango.server.Device`::
 
-        from tango.server import Device, DeviceMeta, run
+        from tango.server import Device, run
 
         class PowerSupply(Device):
             pass
@@ -1424,7 +1441,7 @@ def run(classes, args=None, msg_stream=sys.stdout,
     classes `MyServerClass` and `MyServer`::
 
         from tango import Device_4Impl, DeviceClass
-        from tango.server import Device, DeviceMeta, run
+        from tango.server import Device, run
 
         class PowerSupply(Device):
             pass
